"""
Provides classes for the Merkle-tree's leaves and internal nodes
"""

from abc import ABCMeta, abstractmethod

from pymerkle.serializers import NodeSerializer, LeafSerializer
from pymerkle.exceptions import (NoChildException, NoDescendantException,
    NoParentException, LeafConstructionError, UndecodableArgumentError,
    UndecodableRecord)
from pymerkle.utils import NONE
import json


# Prefices used for node and tree printing

L_BRACKET_SHORT = '\u2514' + '\u2500'           # └─
L_BRACKET_LONG  = '\u2514' + 2 * '\u2500'       # └──
T_BRACKET       = '\u251C' + 2 * '\u2500'       # ├──
VERTICAL_BAR    = '\u2502'                      # │


class __Node(object, metaclass=ABCMeta):
    """
    Abstract base class for Merkle-tree leaves and internal nodes
    """

    __slots__ = ('__encoding', '__child',)

    def __init__(self, encoding):
        self.__encoding = encoding


    @abstractmethod
    def serialize(self):
        """
        """

    @abstractmethod
    def toJSONString(self):
        """
        """

    @property
    def encoding(self):
        return self.__encoding


    @property
    def child(self):
        """
        :raises NoChildException: if the node has no *.child* attribute
        """
        try:
            return self.__child
        except AttributeError:
            raise NoChildException

    def set_child(self, child):
        self.__child = child


    @property
    def left(self):
        """
        :raises NoChildException: if the node has no *.left* attribute
        """
        try:
            return self.__left
        except AttributeError:
            raise NoParentException


    @property
    def right(self):
        """
        :raises NoChildException: if the node has no *.right* attribute
        """
        try:
            return self.__right
        except AttributeError:
            raise NoParentException


    def is_left_parent(self):
        """
        Checks if the node is a left parent.

        :returns: *True* iff the node is the *.left* attribute of some
                other node inside the containing tree
        :rtype: bool
        """
        try:
            _child = self.child
        except NoChildException:
            return False
        return self == _child.left


    def is_right_parent(self):
        """
        Checks if the node is a right parent.

        :returns: *True* iff the node is the *.right* attribute of some
                other node inside the containing tree
        :rtype: bool
        """
        try:
            _child = self.child
        except NoChildException:
            return False
        return self == _child.right


    def is_parent(self):
        """
        Checks if the node is a parent.

        :returns: *True* iff the node is the *.right* or *.left*
            attribute of some other node inside the containing tree
        :rtype: bool
        """
        try:
            self.child
        except NoChildException:
            return False
        return True


    def descendant(self, degree):
        """
        Detects and returns the node that is *degree* steps
        upwards within the containing Merkle-tree.

        .. note:: Descendant of degree 0 is the node itself, descendant
                of degree 1 is the node's child, etc.

        :param degree: depth of descendancy
        :type degree:  int
        :returns:      the descendant corresdponding to the requested depth
        :rtype:        __Node

        :raises NoDescendantException: if the provided degree
            exceeds possibilities
        """
        if degree == 0:
            return self
        else:
            try:
                _child = self.child
            except NoChildException:
                raise NoDescendantException
            return _child.descendant(degree - 1)


    def __repr__(self):
        """
        Overrides the default implementation.

        Sole purpose of this function is to easy display info
        about the node by just invoking it at console.

        .. warning:: Contrary to convention, the output of this implementation
            is not insertible into the *eval()* builtin Python function
        """
        def memory_id(obj): return str(hex(id(obj)))
        try:
            child_id = memory_id(self.child)
        except NoChildException:
            child_id = NONE
        try:
            left_id  = memory_id(self.left)
        except NoParentException:
            left_id  = NONE
            right_id = NONE
        else:
            right_id = memory_id(self.right)

        return '\n    memory-id    : {self_id}\
                \n    left parent  : {left_id}\
                \n    right parent : {right_id}\
                \n    child        : {child_id}\
                \n    hash         : {hash}\n'\
                .format(self_id=memory_id(self),
                    left_id=left_id,
                    right_id=right_id,
                    child_id=child_id,
                    hash=self.digest.decode(self.encoding))


    def __str__(self, encoding=None, level=0, indent=3, ignore=[]):
        """
        Overrides the default implementation.

        Designed so that inserting the node as an argument to the builtin
        *print()* Python function displays the subtree of the Merkle-tree
        whose root is the present node.

        Sole purpose of this function is to be used for printing Merkle-trees
        in a terminal friendly way (similar to what is printed at console when
        running the ``tree`` command of Unix based platforms)

        :param encoding: [optional] encoding type to be used for decoding
                    the digest stored by the present node
        :type encoding: str
        :param level: [optional] Defaults to 0. Must be left equal to the
                default value when called externally by the user. Increased by
                1 whenever the function is recursively called, in order for
                track be kept of depth while printing
        :type level: int
        :param indent: [optional] Defaults to 3. The horizontal depth at
                    which each level of the tree will be indented with
                    respect to the previous one. Increase to achieve
                    better visibility of the tree's structure.
        :type indent: int
        :param ignore: [optional] Defaults to the empty list. Must be left
                    equal to the *default* value when called externally by the
                    user. Augmented appropriately whenever the function is
                    recursively invoked, in order for track to be kept of the
                    positions where vertical bars should be omitted.
        :type ignore: list of integers
        :rtype: str

        .. note:: Left parents appear above the right ones.
        """
        if level == 0:
            output = '\n'
            if not self.is_left_parent() and not self.is_right_parent(): # root case
                output += f' {L_BRACKET_SHORT}'
        else:
            output = (indent + 1) * ' '
        for _ in range(1, level):
            if _ not in ignore:
                output += f' {VERTICAL_BAR}'              # Include vertical bar
            else:
                output += 2 * ' '
            output += indent * ' '
        new_ignore = ignore[:]
        del ignore
        if self.is_left_parent():
            output += f' {T_BRACKET}'
        if self.is_right_parent():
            output += f' {L_BRACKET_LONG}'
            new_ignore.append(level)
        encoding = encoding if encoding else self.encoding
        output += f'{self.digest.decode(encoding)}\n'
        if not isinstance(self, Leaf):                          # Recursive step
            output += self.left.__str__(encoding, level + 1,
                                indent, new_ignore)
            output += self.right.__str__(encoding, level + 1,
                                indent, new_ignore)
        return output


class Leaf(__Node):
    """
    Class for the Merkle-tree's leaves

    By leaf is meant a parentless node storing the checksum
    of some encrypted record

    :param hash_func: hash function to be used for encryption.
    :type hash_func: method
    :param encoding: encoding type to be used when decoding the
            digest stored by the leaf
    :type encoding: str
    :param record: [optional] the record to be encrypted within the leaf.
            If provided, then *digest* should not be provided.
    :type record: str or bytes
    :param digest: [optional] The checksum to be stored by the leaf.
                If provided, then *record* should not be provided.
    :type digest: str

    :raises LeafConstructionError: if both *record* and *digest* were
        provided
    """

    __slots__ = ('__digest',)

    def __init__(self, hash_func, encoding, record=None, digest=None):
        if digest is None and record:
            try:
                digest = hash_func(record)
            except UndecodableArgumentError:
                raise UndecodableRecord
            else:
                super().__init__(encoding)
                self.__digest = digest
        elif record is None and digest:
            super().__init__(encoding)
            self.__digest = bytes(digest, encoding)
        else:
            err = 'Either record or digest may be provided'
            raise LeafConstructionError(err)

    @property
    def digest(self):
        """
        The checksum currently stored by the leaf.

        :rtype: bytes
        """
        return self.__digest

    def serialize(self):
        """
        Returns a JSON entity with the leaf's characteristics as key-value pairs.

        :rtype: dict
        """
        return LeafSerializer().default(self)


    def toJSONString(self):
        """
        Returns a JSON text with the leaf's characteristics as key-value pairs.

        :rtype: str
        """
        return json.dumps(self, cls=LeafSerializer, sort_keys=True, indent=4)


class Node(__Node):
    """
    Class for Merkle-tree's internal nodes

    By internal is meant a node with exactly two parents.

    :param hash_func: hash function to be used for encryption
    :type hash_func: method
    :param encoding: encoding type to be used when decoding the digest
            stored by the node
    :type encoding: str
    :param left: [optional] the node's left parent
    :type left: __Node
    :param right: [optional] the node's right parent
    :type right: __Node
    """

    __slots__ = ('__digest', '__left', '__right',)

    def __init__(self, hash_func, encoding, left, right):
        super().__init__(encoding=encoding)

        digest = hash_func(left.digest, right.digest)
        self.__digest = digest
        self.__left   = left
        self.__right  = right
        left.__child  = self
        right.__child = self

    @property
    def digest(self):
        """
        The checksum currently stored by the node.

        :rtype: bytes
        """
        return self.__digest

    def set_right(self, right):
        """
        Sets the node's right parent.

        :param right: the new right parent
        :type: __Node
        """
        self.__right = right

    def recalculate_hash(self, hash_func):
        """
        Recalculates the node's digest under account of the (possibly new)
        digests stored by its parents.

        :param hash_func: hash function to be used for recalculation
        :type hash_func: method
        """
        self.__digest = hash_func(self.left.digest, self.right.digest)


    def serialize(self):
        """
        Returns a JSON entity with the node's characteristics as key-value pairs.

        :rtype: dict

        .. note:: The *.child* attribute is ommited from node serialization
            in order for circular reference error to be avoided.
        """
        return NodeSerializer().default(self)

    def toJSONString(self):
        """
        Returns a JSON text with the node's characteristics as key-value pairs.

        :rtype: str
        """
        return json.dumps(self, cls=NodeSerializer, sort_keys=True, indent=4)



"""
Provides high-level prover interface for Merkle-trees
"""

from abc import ABCMeta, abstractmethod
import uuid
from time import time, ctime
import json
from pymerkle.exceptions import (NoPathException, InvalidChallengeError,)
from pymerkle.serializers import ProofSerializer
from pymerkle.utils import stringify_path


class Prover(object, metaclass=ABCMeta):
    """
    High-level prover interface for Merkle-trees
    """

    @property
    @abstractmethod
    def length(self):
        """
        """

    @abstractmethod
    def find_index(self, checksum):
        """
        """

    @abstractmethod
    def audit_path(self, index):
        """
        """

    @abstractmethod
    def multi_hash(self, signed_hashes, start):
        """
        """

    @abstractmethod
    def consistency_path(self, sublength):
        """
        """

    @abstractmethod
    def get_commitment(self):
        """
        """

    def merkleProof(self, challenge, commit=True):
        """
        Response of the Merkle-tree to the request of providing a
        Merkle-proof based upon the provided challenge

        :type challenge: dict
        :rtype: Proof

        .. warning:: Provided challenge must be of the form

            ``{'checksum': <str> or <bytes>}`` or ``{'subhash': <str> or <bytes>}``,

            otherwise an ``InvalidChallengeError`` is raised.
        """
        keys = set(challenge.keys())
        if keys == {'checksum'}:
            checksum = challenge['checksum']
            return self.auditProof(checksum, commit=commit)
        elif keys == {'subhash'}:
            subhash = challenge['subhash']
            return self.consistencyProof(subhash, commit=commit)
        raise InvalidChallengeError


    def auditProof(self, checksum, commit=False):
        """
        Response of the Merkle-tree to the request of providing an
        audit proof based upon the provided checksum

        :param checksum: Checksum which the requested proof is to be based upon
        :type checksum: str or bytes
        :rtype: Proof

        :raises InvalidChallengeError: if the provided argument's type
            is not as prescribed
        """
        if isinstance(checksum, str):
            checksum = checksum.encode()
        elif not isinstance(checksum, bytes):
            raise InvalidChallengeError

        index = self.find_index(checksum)
        commitment = self.get_commitment() if commit else None

        try:
            proof_index, audit_path = self.audit_path(index)
        except NoPathException:
            proof = Proof(
                provider=self.uuid,
                hash_type=self.hash_type,
                encoding=self.encoding,
                security=self.security,
                raw_bytes=self.raw_bytes,
                commitment=commitment,
                proof_index=-1,
                proof_path=())
        else:
            proof = Proof(
                provider=self.uuid,
                hash_type=self.hash_type,
                encoding=self.encoding,
                security=self.security,
                raw_bytes=self.raw_bytes,
                commitment=commitment,
                proof_index=proof_index,
                proof_path=audit_path)

        return proof


    def consistencyProof(self, subhash, commit=False):
        """
        Response of the Merkle-tree to the request of providing a consistency
        proof for the acclaimed root-hash of some previous state

        :param subhash: acclaimed root-hash of some previous
                state of the Merkle-tree
        :type subhash: str or bytes
        :type subhash: bytes
        :rtype: Proof

        :raises InvalidChallengeError: if type of *subhash* is not as prescribed
        """
        if isinstance(subhash, str):
            subhash = subhash.encode()
        elif not isinstance(subhash, bytes):
            raise InvalidChallengeError

        commitment = self.get_commitment() if commit is True else None

        proof = Proof(
            provider=self.uuid,
            hash_type=self.hash_type,
            encoding=self.encoding,
            raw_bytes=self.raw_bytes,
            security=self.security,
            commitment=commitment,
            proof_index=-1,
            proof_path=())

        for sublength in range(1, self.length + 1):
            try:
                proof_index, left_path, full_path = self.consistency_path(sublength)
            except NoPathException:
                pass
            else:
                if subhash == self.multi_hash(left_path, len(left_path) - 1):
                    proof = Proof(
                        provider=self.uuid,
                        hash_type=self.hash_type,
                        encoding=self.encoding,
                        raw_bytes=self.raw_bytes,
                        security=self.security,
                        commitment=commitment,
                        proof_index=proof_index,
                        proof_path=full_path)
                    break

        return proof


class Proof(object):
    """
    Class for Merkle-proofs

    :param provider: uuid of the provider Merkle-tree
    :type provider: str
    :param hash_type: hash type of the provider Merkle-tree
    :type hash_type: str
    :param encoding: encoding type of the provider Merkle-tree
    :type encoding: str
    :param raw_bytes: raw-bytes mode of the provider Merkle-tree
    :type raw_bytes: bool
    :param security: security mode of the provider Merkle-tree
    :type security: bool
    :param proof_index: starting position of subsequent validation procedure
    :type proof_index: int
    :param proof_path: path of signed hashes
    :type proof_path: tuple of (+1/-1, bytes)

    Proofs are meant to be output of proof generation mechanisms and not
    manually constructed. Proof construction via deserialization might though
    have practical importance, so that given a proof *p* the following
    constructions are possible:

    >>> from pymerkle import Proof
    >>>
    >>> q = Proof(from_dict=p.serialize())
    >>> r = Proof(from_json=p.toJSONtext())

    or, more uniformly,

    >>> q = Proof.deserialize(p.serialize())
    >>> r = Proof.deserialize(p.toJSONtext())

    .. note:: This is a genuine replication, since deserializations will have
        the same uuid and timestamp as the original.

    :ivar header: (*dict*) contains the keys *uuid*, *timestamp*,
        *creation_moment*, *provider*, *hash_type*, *encoding*,
        *raw_bytes*, *security* and *status*
    :ivar body: (*dict*) Contains the keys *proof_index* and *proof_path*
    """

    def __init__(self, **kwargs):
        """
        """
        header = {}
        body = {}
        if kwargs.get('from_dict'):                             # from json dict
            input = kwargs['from_dict']
            header.update(input['header'])
            if header['commitment']:
                header['commitment'] = header['commitment'].encode()
            body['proof_index'] = input['body']['proof_index']
            body['proof_path'] = tuple((
                pair[0],
                bytes(pair[1], header['encoding'])
            ) for pair in input['body']['proof_path'])
        elif kwargs.get('from_json'):                           # from json text
            input = json.loads(kwargs['from_json'])
            header.update(input['header'])
            if header['commitment']:
                header['commitment'] = header['commitment'].encode()
            body['proof_index'] = input['body']['proof_index']
            body['proof_path'] = tuple((
                pair[0],
                bytes(pair[1], header['encoding'])
            ) for pair in input['body']['proof_path'])
        else:                                                  # multiple kwargs
            header.update({
                'uuid': str(uuid.uuid1()),
                'timestamp': int(time()),
                'creation_moment': ctime(),
                'provider': kwargs['provider'],
                'hash_type': kwargs['hash_type'],
                'encoding': kwargs['encoding'],
                'raw_bytes': kwargs['raw_bytes'],
                'security': kwargs['security'],
                'commitment': kwargs.get('commitment'),
                'status': None})
            body.update({
                'proof_index': kwargs['proof_index'],
                'proof_path': kwargs['proof_path']})
        self.header = header
        self.body = body


    @classmethod
    def deserialize(cls, serialized):
        """
        Deserializes the provided JSON entity

        :params serialized: a Python dict or JSON text, assumed to be the
            serialization of a *Proof* object
        :type: dict or str
        :rtype: Proof
        """
        kwargs = {}
        if isinstance(serialized, dict):
            kwargs.update({'from_dict': serialized})
        elif isinstance(serialized, str):
            kwargs.update({'from_json': serialized})
        return cls(**kwargs)


    def get_validation_params(self):
        """
        Extracts from the proof's header the fields required for configuring
        correctly the validator's hashing machinery.

        :rtype: dict
        """
        header = self.header
        validation_params = dict({
            'hash_type': header['hash_type'],
            'encoding': header['encoding'],
            'raw_bytes': header['raw_bytes'],
            'security': header['security'],
        })
        return validation_params


    def __repr__(self):
        """
        Overrides the default implementation.

        Sole purpose of this function is to display info
        about a proof by just invoking it at console

        .. warning:: Contrary to convention, the output of this implementation
            is not insertible into the *eval()* builtin Python function
        """
        header = self.header
        body = self.body
        encoding = header['encoding']

        return '\n    ----------------------------------- PROOF ------------------------------------\
                \n\
                \n    uuid        : {uuid}\
                \n\
                \n    timestamp   : {timestamp} ({creation_moment})\
                \n    provider    : {provider}\
                \n\
                \n    hash-type   : {hash_type}\
                \n    encoding    : {encoding}\
                \n    raw_bytes   : {raw_bytes}\
                \n    security    : {security}\
                \n\
                \n    proof-index : {proof_index}\
                \n    proof-path  :\
                \n    {proof_path}\
                \n\
                \n    commitment  : {commitment}\
                \n\
                \n    status      : {status}\
                \n\
                \n    -------------------------------- END OF PROOF --------------------------------\
                \n'.format(
                    uuid=header['uuid'],
                    timestamp=header['timestamp'],
                    creation_moment=header['creation_moment'],
                    provider=header['provider'],
                    hash_type=header['hash_type'].upper().replace('_', '-'),
                    encoding=header['encoding'].upper().replace('_', '-'),
                    raw_bytes='TRUE' if header['raw_bytes'] else 'FALSE',
                    security='ACTIVATED' if header['security'] else 'DEACTIVATED',
                    commitment=header['commitment'].decode() \
                    if header['commitment'] else None,
                    proof_index=body['proof_index'],
                    proof_path=stringify_path(body['proof_path'], header['encoding']),
                    status='UNVALIDATED' if header['status'] is None \
                    else 'VALID' if header['status'] is True else 'NON VALID')


# Serialization

    def serialize(self):
        """
        Returns a JSON entity with the proof's characteristics
        as key-value pairs.

        :rtype: dict
        """
        return ProofSerializer().default(self)

    def toJSONString(self):
        """
        Returns a JSON text with the proof's characteristics
        as key-value pairs.

        :rtype: str
        """
        return json.dumps(self, cls=ProofSerializer, sort_keys=True, indent=4)



"""
Provides high-level encryption interface for Merkle-trees
"""

from abc import ABCMeta, abstractmethod
import os
import json
import mmap
import contextlib
from tqdm import tqdm

from pymerkle.exceptions import UndecodableRecord

abspath = os.path.abspath

class Encryptor(object, metaclass=ABCMeta):
    """
    High-level encryption interface for Merkle-trees
    """

    @abstractmethod
    def update(self, record):
        """
        """

    def encryptRecord(self, record):
        """
        Updates the Merkle-tree by storing the checksum of the provided record
        into a newly-created leaf.

        :param record: Record whose checksum is to be stored into a new leaf
        :type record: str or bytes

        :raises UndecodableRecord: if the tree does not accept arbitrary bytes
            and the provided record is out of its configured encoding type
        """
        try:
            self.update(record=record)
        except UndecodableRecord:
            raise


    def encryptFileContent(self, file_path):
        """
        Encrypts the provided file as a single new leaf into the Merkle-tree.

        Updates the Merkle-tree with *one* newly-created leaf storing the
        checksum of the provided file's content.

        :param file_path: Relative path of the file under encryption with
                respect to the current working directory
        :type file_path: str

        :raises UndecodableRecord: if the tree does not accept arbitrary bytes
            and the provided files contains sequences out of the tree's
            configured encoding type
        """
        with open(abspath(file_path), mode='r') as __file:
            with contextlib.closing(
                mmap.mmap(
                    __file.fileno(),
                    0,
                    access=mmap.ACCESS_READ
                )
            ) as __buffer:
                try:
                    self.update(record=__buffer.read())
                except UndecodableRecord:
                    raise


    def encryptFilePerLog(self, file_path):
        """
        Per log encryption of the provided file into the Merkle-tree.

        Successively updates the tree with each line of the provided
        file in respective order

        :param file_path: Relative path of the file under enryption with
            respect to the current working directory
        :type file_path: str

        :raises UndecodableRecord: if the tree does not accept arbitrary bytes
            and the provided files contains sequences out of the tree's
            configured encoding type
        """
        absolute_file_path = abspath(file_path)
        with open(absolute_file_path, mode='r') as __file:
            buffer = mmap.mmap(
                __file.fileno(),
                0,
                access=mmap.ACCESS_READ
            )

        # Extract logs
        records = []
        readline = buffer.readline
        append = records.append
        if not self.raw_bytes:
            # ~ Check that no line of the provided file is outside
            # ~ the tree's encoding type and discard otherwise
            encoding = self.encoding
            while 1:
                record = readline()
                if not record:
                    break
                try:
                    record = record.decode(encoding)
                except UnicodeDecodeError as err:
                    raise UndecodableRecord(err)
                append(record)
        else:
            # ~ No need to check anything, just load all lines
            while 1:
                record = readline()
                if not record:
                    break
                append(record)

        # Perform line by line encryption
        tqdm.write('')
        update = self.update
        for record in tqdm(records, desc='Encrypting file per log', total=len(records)):
            update(record=record)
        tqdm.write('Encryption complete\n')


    def encryptJSON(self, object, sort_keys=False, indent=0):
        """
        Encrypts the provided JSON entity as a single new leaf into the
        Merkle-tree.

        Updates tree with *one* newly-created leaf storing the checksum of the
        provided object's stringification.

        :param object: JSON entity under encryption
        :type objec: dict
        :param sort_keys: [optional] Defaults to *False*. If *True*, then
            the object's keys are alphabetically sorted before its
            stringification.
        :type sort_keys: bool
        :param indent: [optional] Defaults to 0. Specifies key indentation
            upon stringification of the provided JSON.
        :type indent: int
        """
        self.update(
            record=json.dumps(object, sort_keys=sort_keys, indent=indent))


    def encryptJSONFromFile(self, file_path, sort_keys=False, indent=0):
        """
        Encrypts the object from within the provided *.json* file as a
        single new leaf into the Merkle-tree.

        Updates the tree with *one* newly-created leaf storing the checksum of
        the provided JSON's stringification.

        :param file_path: Relative path of a *.json* file with respect to the
            current working directory
        :type file_path: str
        :param sort_keys: [optional] Defaults to *False*. If *True*, then
            the object's keys are alphabetically sorted before its
            stringification
        :type sort_keys: bool
        :param indent: [optional] Defaults to 0. Specifies key indentation
                upon stringification of the object under encryption
        :type indent: sint

        :raises JSONDecodeError: if the specified file could not be deserialized
        """
        try:
            with open(abspath(file_path), 'rb') as __file:
                object = json.load(__file)
        except json.JSONDecodeError:
            raise
        record = json.dumps(object, sort_keys=sort_keys, indent=indent)
        self.update(record=record)



"""
Provides the class for Merkle-trees containing the low-level algorithms
of proof generation
"""

from .encryption import Encryptor
from .prover import Prover
from .nodes import Node, Leaf
from pymerkle.hashing import HashMachine
from pymerkle.serializers import MerkleTreeSerializer
from pymerkle.utils import log_2, decompose, NONE
from pymerkle.exceptions import (LeafConstructionError, NoChildException,
    EmptyTreeException, NoPathException, InvalidTypes, NoSubtreeException,
    NoPrincipalSubroots, InvalidComparison, WrongJSONFormat, UndecodableRecord)
import uuid
import json
from tqdm import tqdm

NONE_BAR = '\n ' + '\u2514' + '\u2500' + NONE  # └─[None]


class MerkleTree(HashMachine, Encryptor, Prover):
    """
    Class for Merkle-trees

    :param \*records: [optional] Records encrypted into the Merkle-tree at
                construction.
    :type \*records: str or bytes
    :param hash_type: [optional] Specifies the Merkle-tree's hashing algorithm.
                    Defaults to *sha256*.
    :type hash_type: str
    :param encoding: [optional] Specifies the Merkle-tree's encoding type.
                    Defaults to *utf_8*.
    :type encoding: str
    :param raw_bytes: [optional] Specifies whether the Merkle-tree will accept
                raw binary data (independently of its configured encoding
                type). Defaults to *True*.
    :type raw_bytes: bool
    :param security: [optional Specifies if defense against second-preimage
                attack is enabled. Defaults to *True*.
    :type security: bool

    :ivar uuid: (*str*) uuid of the Merkle-tree (time-based)
    :ivar hash_type: (*str*) See the constructor's homonymous argument
    :ivar encoding: (*str*) See the constructor's homonymous argument
    :ivar raw_bytes: (*bool*) See the constructor's homonymous argument
    :ivar security: (*bool*) See the constructor's homonymous argument
    """

    def __init__(self, *records, hash_type='sha256', encoding='utf-8',
            raw_bytes=True, security=True):
        self.uuid = str(uuid.uuid1())

        # Hash-machine configuration
        super().__init__(hash_type, encoding, raw_bytes, security)

        # Tree generation
        self.leaves = []
        self.nodes  = set()
        for record in records:
            try:
                self.update(record=record)
            except UndecodableRecord:
                raise

    def clear(self):
        """
        Deletes all nodes of the Merkle-tree
        """
        self.leaves = []
        self.nodes = set()
        try:
            del self.__root
        except AttributeError:
            pass

    def __bool__(self):
        """
        :returns: *False* iff the Merkle-tree is empty (no nodes)
        :rtype: bool
        """
        return bool(self.nodes)

    @property
    def root(self):
        """
        Current root of the Merkle-tree

        :returns: The tree's current root-node
        :rtype: __Node

        :raises EmptyTreeException: if the Merkle-tree is currently empty
        """
        if not self:
            raise EmptyTreeException
        return self.__root

    @property
    def rootHash(self):
        """
        :returns: Current root-hash of the Merkle-tree
        :rtype:   bytes

        :raises EmptyTreeException: if the Merkle-tree is currently empty
        """
        try:
            root = self.__root
        except AttributeError:
            raise EmptyTreeException
        return root.digest

    def get_commitment(self):
        """
        Returns the current root-hash of the Merkle-tree if the latter is
        not empty, otherwise *None*.

        :rtype: bytes or None
        """
        commitment = None
        try:
            commitment = self.rootHash
        except EmptyTreeException:
            pass
        return commitment

    @property
    def length(self):
        """
        Current length of the Merkle-tree (i.e., number of its leaves)

        :rtype: int
        """
        return len(self.leaves)

    @property
    def size(self):
        """
        Current number of the Merkle-tree's nodes

        :rtype: int
        """
        return len(self.nodes)

    @property
    def height(self):
        """
        Current height of the Merkle-tree

        .. note:: Since the tree is binary *balanced*, its height coincides
                with the length of its leftmost branch

        :rtype: int
        """
        length = len(self.leaves)
        if length == 0:
            return 0
        return log_2(length) + 1 if length != 2 ** log_2(length) \
            else log_2(length)


    def update(self, record=None, digest=None):
        """
        Updates the Merkle-tree by storing the digest of the inserted record
        into a newly-created leaf. Restructures the tree appropriately and
        recalculates appropriate interior hashes

        :param record: [optional] The record whose digest is to be stored into
                    a new leaf
        :type record:  str or bytes
        :param digest: [optional] The digest to be stored into the new leaf
        :type digest:  str

        .. warning:: Exactly one of either record or digest
            should be provided

        :raises LeafConstructionError: if both record and digest
                                    were provided
        :raises UndecodableRecord: if the Merkle-tree is not in raw-bytes mode
            and the provided record does not fall under its configured type
        """
        encoding = self.encoding
        hash = self.hash

        if self:
            leaves = self.leaves
            append = leaves.append
            add = self.nodes.add

            # ~ Height and root of the *full* binary subtree with maximum
            # ~ possible length containing the rightmost leaf
            last_power   = decompose(len(leaves))[-1]
            last_subroot = leaves[-1].descendant(degree=last_power)

            # Encrypt new record into new leaf
            try:
                new_leaf = Leaf(hash, encoding, record, digest)
            except (LeafConstructionError, UndecodableRecord):
                raise

            # Assimilate new leaf
            append(new_leaf)
            add(new_leaf)
            try:
                old_child = last_subroot.child                                  # Save child info before bifurcation
            except NoChildException:                                            # last_subroot was previously root
                self.__root = Node(hash, encoding, last_subroot, new_leaf)
                add(self.__root)
            else:
                # Create bifurcation node
                new_child = Node(hash, encoding, last_subroot, new_leaf)
                add(new_child)

                # Interject bifurcation node
                old_child.set_right(new_child)
                new_child.set_child(old_child)

                # Recalculate hashes only at the rightmost branch of the tree
                current_node = old_child
                while 1:
                    current_node.recalculate_hash(hash_func=hash)
                    try:
                        current_node = current_node.child
                    except NoChildException:
                        break
        else:                                                                   # Empty tree case
            try:
                new_leaf = Leaf(hash, encoding, record, digest)
            except (LeafConstructionError, UndecodableRecord):
                raise
            self.leaves = [new_leaf]
            self.nodes = set([new_leaf])
            self.__root = new_leaf


    # Low-level audit proof

    def audit_path(self, index):
        """
        Low-level audit proof

        Computes and returns the audit-path corresponding to the provided leaf
        index along with the position where subsequent proof validation should
        start from.

        :param index: position (zero based leaf index) where audit-path
                computation should be based upon
        :type index: int
        :returns: Starting position of subsequent proof validation along with
            a sequence of signed checksums (the sign +1 or -1 indicating
            pairing with the right or left neighbour respectively)
        :rtype: (int, tuple of (+1/-1, bytes))

        :raises NoPathException: if the provided index exceed's the tree's
            current length
        """
        if index < 0:
            # ~ Handle negative index case as NoPathException, since
            # ~ certain negative indices might otherwise be
            # ~ considered as valid positions
            raise NoPathException

        try:
            current_node = self.leaves[index]
        except IndexError:
            raise NoPathException # Covers also the empty tree case

        initial_sign = +1
        if current_node.is_right_parent():
            initial_sign = -1
        path = [(initial_sign, current_node.digest)]
        append = path.append
        start = 0
        while 1:
            try:
                current_child = current_node.child
            except NoChildException:
                break
            if current_node.is_left_parent():
                next_digest = current_child.right.digest
                if current_child.is_left_parent():
                    append((+1, next_digest))
                else:
                    append((-1, next_digest))
            else:
                next_digest = current_child.left.digest
                if current_child.is_right_parent():
                    path.insert(0, (-1, next_digest))
                else:
                    path.insert(0, (+1, next_digest))
                start += 1
            current_node = current_child
        return start, tuple(path)


    def find_index(self, checksum):
        """
        Detects the (zero-based) index of the leftmost leaf which stores the
        provided checksum

        :type: bytes
        """
        # ~ Detect the index of the first leaf storing the provided checksum;
        # ~ if no such leaf exists (i.e., the inserted argument has not been
        # ~ encrypted into the tree), leave index equal to -1 so that an
        # ~ appropriate NoPathException be subsequently raised
        index = -1
        count = 0
        leaves = (leaf for leaf in self.leaves)
        while 1:
            try:
                leaf = next(leaves)
            except StopIteration:
                break
            if checksum == leaf.digest:
                index = count
                break
            count += 1
        return index


    # Low-level consistency proof

    def consistency_path(self, sublength):
        """
        Low-level consistency proof

        Computes and returns the consistency-path corresponding to the tree's
        length for a previous state, along with the position where subsequent
        proof validation should start from and the sequence of subroots
        constituting the produced path from the left.

        :param sublength: any number equal to or smaller than the tree's
                    current length
        :type sublength: int
        :returns: Starting position of subsequent proof validation along with
            sequence of subroots constituting the produced path from the left
            and the path of signed hashes per se (the sign +1 or -1 indicating
            pairing with the right or left neighbour respectively)
        :rtype: (int, tuple of (+1/-1, bytes), tuple of (+1/-1, bytes))

        :raises NoPathException: if the provided *sublength* is non-positive
            or no sequence of subroots corresponds to it
        """
        if sublength < 0 or self.length == 0:
            raise NoPathException
        try:
            left_subroots = self.principal_subroots(sublength)
        except NoPrincipalSubroots:
            raise NoPathException                                           # Incompatibility issue detected

        right_subroots = self.minimal_complement(left_subroots)
        all_subroots = left_subroots + right_subroots
        if not right_subroots or not left_subroots:
            all_subroots = [(-1, _[1]) for _ in all_subroots]               # Reset all signs to minus
            proof_index = len(all_subroots) - 1                             # Will start multi-hashing from endpoint
        else:
            proof_index = len(left_subroots) - 1                            # Will start multi-hashing from midpoint

        # Collect sign-hash pairs
        left_path = tuple((-1, _[1].digest) for _ in left_subroots)
        full_path = tuple((_[0], _[1].digest) for _ in all_subroots)

        return proof_index, left_path, full_path

    def minimal_complement(self, subroots):
        """
        Complements optimally from the right the provided sequence of subroots,
        so that a full consistency-path be subsequently generated.

        :param subroots: roots of a complete leftomost sequence of
                full binary subtrees
        :type subroots: list of nodes
        :rtype: list of (+1/-1, bytes)
        """
        if len(subroots) == 0:
            return self.principal_subroots(self.length)

        complement = []
        append = complement.append
        while 1:
            try:
                subroots[-1][1].child
            except NoChildException:
                break

            subroot = subroots[-1][1]
            if subroot.is_left_parent():
                if subroot.child.is_right_parent():
                    append((-1, subroot.child.right))
                else:
                    append((+1, subroot.child.right))
                subroots = subroots[:-1]
            else:
                subroots = subroots[:-2]
            subroots.append((+1, subroot.child))

        return complement

    def principal_subroots(self, sublength):
        """
        Detects in corresponding order the roots of the successive, leftmost,
        full binary subtrees of maximum (and thus decreasing) length, whose
        lengths sum up to the provided argument. Detected nodes are prepended
        with a sign (+1 or -1), carrying information for subsequent generation
        of consistency proofs.

        :param sublength: non negative integer smaller than or equal to the
                tree's current length, such that the corresponding sequence
                of subroots exists
        :returns: Signed roots of the detected subtrees, whose hashes to be
                    utilized in generation of consistency proofs
        :rtype: list of signed nodes

        :raises NoPrincipalSubroots: if the provided number does not fulfill
            the prescribed conditions
        """
        if sublength < 0:
            raise NoPrincipalSubroots                                  # Mask negative input case as incompatibility

        principal_subroots = []
        append = principal_subroots.append
        powers = decompose(sublength)
        start = 0
        for power in powers:
            try:
                subroot = self.subroot(start, power)
            except NoSubtreeException:
                raise NoPrincipalSubroots                              # Incompatibility issue detected

            try:
                child = subroot.child
                grandchild = child.child
            except NoChildException:
                if subroot.is_left_parent():
                    append((+1, subroot))
                else:
                    append((-1, subroot))
            else:
                if child.is_left_parent():
                    append((+1, subroot))
                else:
                    append((-1, subroot))
            finally:
                start += 2 ** power

        if len(principal_subroots) > 0:
            principal_subroots[-1] = (+1, principal_subroots[-1][1])    # Modify last sign
        return principal_subroots

    def subroot(self, start, height):
        """
        Detects the root of the unique full binary subtree with leftmost
        leaf located at position *start* and height equal to *height*.

        :param start: leaf position (zero based) where detection of
                subtree should start from
        :type start: int
        :param height: height of candidate subtree to be detected
        :type height: int
        :returns: Root of the detected subtree
        :rtype: __Node

        :raises NoSubtreeException: if no subtree exists for
                the provided parameters
        """
        # Detect candidate subroot
        try:
            subroot = self.leaves[start]
        except IndexError:
            raise NoSubtreeException
        i = 0
        while i < height:
            try:
                next_node = subroot.child
            except NoChildException:
                raise NoSubtreeException
            if next_node.left is not subroot:
                raise NoSubtreeException
            subroot = subroot.child
            i += 1

        # Verify existence of *full* binary subtree
        right_parent = subroot
        i = 0
        while i < height:
            if isinstance(right_parent, Leaf):
                raise NoSubtreeException
            right_parent = right_parent.right
            i += 1

        return subroot


    # Inclusion test

    def inclusionTest(self, subhash):
        """
        Verifies that the provided parameter corresponds to a valid previous
        state of the Merkle-tree

        :param subhash: acclaimed root-hash of some previous
                state of the Merkle-tree
        :type subhash: bytes
        :rtype: bool

        :raises InvalidTypes: if the type of any of the provided
            arguments is not as prescribed
        """
        if not isinstance(subhash, bytes):
            raise InvalidTypes
        included = False
        multi_hash = self.multi_hash
        for sublength in range(1, self.length + 1):
            left_roots = self.principal_subroots(sublength)
            left_path = tuple((-1, _[1].digest) for _ in left_roots)
            if subhash == multi_hash(left_path, len(left_path) - 1):
                included = True
                break
        return included


    # Persistence

    def export(self, file_path):
        """
        Creates a *.json* file located at the provided path and exports into
        it the rquired minimum, so that the Merkle-tree can be retrieved in
        its current state from that file

        .. note:: If the provided path does not end with *.json*, then this
            extension will be automatically appended to it before exporting

        .. warning:: If a file already exists at the provided path,
                then it will be overwritten

        :param file_path: relative path of the export file with respect to the
                current working directory
        :type file_path: str
        """
        with open(f'{file_path}.json' if not file_path.endswith('.json') \
            else file_path, 'w') as __file:
            json.dump(self.serialize(), __file, indent=4)


    @classmethod
    def loadFromFile(cls, file_path):
        """
        Loads a Merkle-tree from the provided file, the latter being the result
        of an export (cf. the *MerkleTree.export()* method)

        :param file_path: relative path of the file to load from with
                respect to the current working directory
        :type file_path: str
        :returns: The tree loaded from the provided file
        :rtype: MerkleTree

        :raises WrongJSONFormat: if the JSON object loaded from within the
                    provided file is not a Merkle-tree export
        """
        with open(file_path, 'r') as __file:
            loaded_object = json.load(__file)
        try:
            header = loaded_object['header']
            tree = cls(
                hash_type=header['hash_type'],
                encoding=header['encoding'],
                raw_bytes=header['raw_bytes'],
                security=header['security'])
        except KeyError:
            raise WrongJSONFormat

        tqdm.write('\nFile has been loaded')
        update = tree.update
        for hash in tqdm(loaded_object['hashes'], desc='Retrieving tree...'):
            update(digest=hash)
        tqdm.write('Tree has been retrieved')
        return tree


    # Comparison

    def __eq__(self, other):
        """
        Implements the ``==`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
            is not instance of the *MerkleTree* class
        """
        if not isinstance(other, self.__class__):
            raise InvalidComparison
        if not other:
            return not self
        return True if not self else self.rootHash == other.rootHash

    def __ne__(self, other):
        """
        Implements the ``!=`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
            is not instance of the *MerkleTree* class
        """
        if not isinstance(other, self.__class__):
            raise InvalidComparison
        if not other:
            return self.__bool__()
        return True if not self else self.rootHash != other.rootHash

    def __ge__(self, other):
        """
        Implements the ``>=`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
        is not instance of the ``tree.MerkleTree`` class
        """
        if not isinstance(other, self.__class__):
            raise InvalidComparison
        if not other:
            return True
        return False if not self else \
            self.inclusionTest(other.rootHash)

    def __le__(self, other):
        """
        Implements the ``<=`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
            is not instance of the *MerkleTree* class
        """

        if not isinstance(other, self.__class__):
            raise InvalidComparison
        return other.__ge__(self)

    def __gt__(self, other):
        """
        Implements the ``>`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
            is not instance of the *MerkleTree* class
        """
        if not isinstance(other, self.__class__):
            raise InvalidComparison
        if not other:
            return self.__bool__()
        elif not self or self.rootHash == other.rootHash:
            return False
        return self.inclusionTest(other.rootHash)

    def __lt__(self, other):
        """
        Implements the ``<`` operator

        :param other: Merkle-tree to compare with
        :type other: MerkleTree

        :raises InvalidComparison: if compared with an object that
            is not instance of the *MerkleTree* class
        """
        if not isinstance(other, self.__class__):
            raise InvalidComparison
        return other.__gt__(self)


    # Representation

    def __repr__(self):
        """
        Overrides the default implementation

        Sole purpose of this function is to display info about
        the Merkle-treee by just invoking it at console

        .. warning:: Contrary to convention, the output of this implementation
            is not insertible to the eval() builtin Python function.
        """

        return '\n    uuid      : {uuid}\
                \n\
                \n    hash-type : {hash_type}\
                \n    encoding  : {encoding}\
                \n    raw-bytes : {raw_bytes}\
                \n    security  : {security}\
                \n\
                \n    root-hash : {root_hash}\
                \n\
                \n    length    : {length}\
                \n    size      : {size}\
                \n    height    : {height}\n'.format(
                    uuid=self.uuid,
                    hash_type=self.hash_type.upper().replace('_', '-'),
                    encoding=self.encoding.upper().replace('_', '-'),
                    raw_bytes='TRUE' if self.raw_bytes else 'FALSE',
                    security='ACTIVATED' if self.security else 'DEACTIVATED',
                    root_hash=self.rootHash.decode(self.encoding) if self else NONE,
                    length=self.length,
                    size=self.size,
                    height=self.height)

    def __str__(self, indent=3):
        """
        Overrides the default implementation

        Designed so that inserting the Merkle-tree into the *print()* function
        displays it in a terminal friendly way, that is, resembles the output
        of the ``tree`` command at Unix based platforms

        :param indent: [optional] The horizontal depth at which each level will
                be indented with respect to its previous one. Defaults to 3.
        :type indent: int
        :rtype: str

        .. note:: Left parents are printed *above* the right ones
        """
        try:
            root = self.root
        except EmptyTreeException:
            return NONE_BAR
        return root.__str__(indent=indent, encoding=self.encoding)


    # Serialization

    def serialize(self):
        """
        Returns a JSON entity with the Merkle-trees's current characteristics
        and digests stored by its leaves.

        :rtype: dict
        """
        return MerkleTreeSerializer().default(self)

    def toJSONString(self):
        """
        Returns a JSON text with the Merkle-tree's current characteristics
        and digests stored by its leaves.

        :rtype: str
        """
        return json.dumps(self,
            cls=MerkleTreeSerializer, sort_keys=True, indent=4)



"""
Provides hash utilities used accross the *pymerkle* library
"""

import hashlib
from .encoding import Encoder
from pymerkle.exceptions import UnsupportedHashType, EmptyPathException


HASH_TYPES = ['md5', 'sha224', 'sha256', 'sha384', 'sha512',
    'sha3_224', 'sha3_256', 'sha3_384', 'sha3_512',]
"""Supported hash types"""


class HashMachine(Encoder):
    """
    Encapsulates the hash utilities used accross the *pymerkle* library

    :param hash_type: [optional] Specifies the hash algorithm used by the
            machine. Defaults to *sha256*.
    :type hash_type: str
    :param encoding: [optional] Specifies the encoding algorithm used by the
            machine before hashing. Defaults to *utf_8*.
    :type encoding: str
    :param raw_bytes: [optional] Specifies whether the machine accepts raw
            binary data independently of its configured encoding type.
            Defaults to *True*.
    :type raw_bytes: bool
    :param security: [optional] Specifies whether defense against
            second-preimage attack will be enabled. Defaults to *True*.
    :type security: bool

    :raises UnsupportedHashType: if the provided encoding is not
                        contained in *pymerkle.hashing.encoding.ENCODINGS*.
    :raises UnsupportedHashType: if the provided hash-type is not
                        contained in *pymerkle.hashing.machine.HASH_TYPES*.

    :ivar algorithm: (*builtin_function_or_method*) Hash algorithm used
                    by the machine
    :ivar encoding: (*str*) See the constructor's homonymous argument
    :ivar raw_bytes: (*bool*) See the constructor's homonymous argument
    :ivar security: (*bool*) See the constructor's homonymous argument
    """
    def __init__(self, hash_type='sha256', encoding='utf-8',
                    raw_bytes=True, security=True):
        ht = hash_type.lower().replace('-', '_')
        if ht not in HASH_TYPES:
            err = f'Hash type {hash_type} is not supported'
            raise UnsupportedHashType(err)
        self.hash_type = ht
        self.algorithm = getattr(hashlib, self.hash_type)
        super().__init__(encoding=encoding, raw_bytes=raw_bytes, security=security)


    def hash(self, left, right=None):
        """
        Core hash utility

        Computes the digest of the provided arguments' concatenation. If only
        one argument is passed in, then the disgest of this single argument
        is computed.

        :param left: left member of the pair to be hashed
        :type left: str or bytes
        :param right: [optional] right member of the pair to be hashed
        :type right: bytes

        :returns: Digest of the entity occuring after concatenation of
                provided arguments
        :rtype: bytes
        """
        data = self.encode(left, right)
        hexdigest = self.algorithm(data).hexdigest()
        return bytes(hexdigest, self.encoding)


    def multi_hash(self, signed_hashes, start):
        """
        Extended hash utility

        Repeatedly applies the *.hash()* method over the provided iterable of
        signed hashes (signs indicating parenthetization during repetition and
        *start* indicating starting position of application).
        Schematically speaking,

        ``multi_hash([(1, a), (1, b), (-1, c), (-1, d)], 1)``

        is equivalent to

        ``hash(hash(a, hash(b, c)), d)``.

        .. note:: If the provided iterable contains only one element, then this
            single hash is returned (without sign), that is,

                 ``multi_hash(signed_hashes=((+/-1, a)), start=0)``

            equals ``a`` (no hashing over single elements)

        .. warning:: When using this method, make sure that the combination of
                signs corresponds indeed to a valid parenthetization

        :param signed_hashes: a finite sequence of signed hashes
        :type signed_hashes: iterable of (+1/-1, bytes)
        :param start: starting position for application of *.hash()*
        :type start: int
        :rtype: bytes

        :raises EmptyPathException: if the provided sequence was empty
        """
        hash = self.hash

        signed_hashes = list(signed_hashes)
        if signed_hashes == []:
            raise EmptyPathException
        elif len(signed_hashes) == 1:
            return signed_hashes[0][1]

        i = start
        while len(signed_hashes) > 1:
            if signed_hashes[i][0] == +1:           # Pair with the right neighbour
                if i == 0:
                    new_sign = +1
                else:
                    new_sign = signed_hashes[i + 1][0]
                new_hash = hash(
                    signed_hashes[i][1],
                    signed_hashes[i + 1][1])
                move = +1
            else:                                   # Pair with the left neighbour
                new_sign = signed_hashes[i - 1][0]
                new_hash = hash(
                    signed_hashes[i - 1][1],
                    signed_hashes[i][1])
                move = -1
            signed_hashes[i] = (new_sign, new_hash)
            del signed_hashes[i + move]             # Shrink
            if move < 0:
                i -= 1
        return signed_hashes[0][1]



"""
Provides utilities for Merkle-proof validation
"""

from pymerkle.hashing import HashMachine
from pymerkle.core.prover import Proof
from pymerkle.exceptions import InvalidMerkleProof
from pymerkle.serializers import ReceiptSerializer
import uuid
from time import time, ctime
import json
import os


class Validator(HashMachine):
    """
    Encapsulates the low-level utility for Merkle-proof validation

    :param input: [optional] a Merkle-proof or its header
    :type: Proof or dict
    """
    def __init__(self, input=None):
        if input is not None:
            if isinstance(input, Proof):
                self.proof = input
                input = input.get_validation_params()
            self.update(input)


    def update(self, input):
        """
        :param input: a Merkle-proof or its header
        :type input: Proof or dict
        """
        if isinstance(input, Proof):
            config = input.get_validation_params()
            self.proof = input
        else:
            config = input
        try:
            hash_type = config['hash_type']
            encoding = config['encoding']
            raw_bytes = config['raw_bytes']
            security = config['security']
        except KeyError as err:
            err = f'Hash-machine could not be configured: Missing parameter: {err}'
            raise KeyError(err)
        super().__init__(hash_type=hash_type, encoding=encoding,
            raw_bytes=raw_bytes, security=security)


    def run(self, proof=None, target=None):
        """
        Runs Merkle-proof validation

        :raises InvalidMerkleProof: if the proof is found to be invalid

        :param proof: the Merkle-proof under validation
        :type proof: Proof
        :param target: [optional] the hash to be be presumably attained at the
            end of the validation process (i.e., acclaimed current root-hash of
            the Merkle-tree having provided the proof). If not explicitly provided,
            should be included in the given proof as the value of the *commitment*
            field
        :type target: bytes
        """
        if proof is None:
            try:
                proof = self.proof
            except AttributeError:
                err = 'No proof provided for validation'
                raise AssertionError(err)
        if target is None:
            try:
                target = proof.header['commitment']
            except KeyError:
                err = 'No acclaimed root-hash provided'
                raise AssertionError(err)
        proof_index = proof.body['proof_index']
        proof_path  = proof.body['proof_path']
        if proof_index == -1 and proof_path == ():
            raise InvalidMerkleProof
        if target != self.multi_hash(proof_path, proof_index):
            raise InvalidMerkleProof


def validateProof(proof, target=None, get_receipt=False, dirpath=None):
    """
    Core utility for Merkle-proof validation

    Validates the provided proof, modifies the proof's status as *True* or
    *False* accordingly and returns result in the form of a boolean or a
    receipt.

    :param proof: the Merkle-proof under validation
    :type proof: Proof
    :param target: [optional] the hash to be be presumably attained at the
        end of the validation process (i.e., acclaimed current root-hash of
        the Merkle-tree having provided the proof). If not explicitly provided,
        should be included in the given proof as the value of the *commitment*
        field
    :type target: bytes
    :param get_receipt: [optional] Specifies whether a receipt will be
        generated for the performed validation instead of a boolean
    :type get_receipt: bool
    :type dirpath: [optional] Relative path with respect to the current working
        directory of the place where the generated receipt is to be saved (as a
        *.json* file named with the receipt's uuid). If not provided, then the
        receipt will not be automatically saved.
    :param dirpath: str
    :returns: Validation result or receipt
    :rtype: bool or Receipt
    """
    validator = Validator(input=proof.get_validation_params())
    if target is None:
        target = proof.header['commitment']
    try:
        validator.run(proof, target)
    except InvalidMerkleProof:
        result = False
    else:
        result = True
    proof_header = proof.header
    proof_header['status'] = result
    receipt = None
    if get_receipt:
        receipt = Receipt(
            proof_uuid=proof_header['uuid'],
            proof_provider=proof_header['provider'],
            result=result)
        if dirpath:
            receipt_header = receipt.header
            with open(
                os.path.join(dirpath, f"{receipt_header['uuid']}.json"),
                'w'
            ) as __file:
                json.dump(receipt.serialize(), __file, sort_keys=True, indent=4)
    return result if not receipt else receipt


class Receipt(object):
    """
    Receipt for Merkle-proof validations

    :param proof_uuid: uuid of the validated proof
    :type proof_uuid: str
    :param proof_provider: uuid of the tree having provided the proof
    :type proof_provider: str
    :param result: result of validation
    :type result: bool

    Receipts are meant to be output of validation mechanisms and not manually
    constructed. Receipt construction via deserialization might though have
    practical importance, so that given a receipt *r* the following
    constructions are possible:

    >>> from pymerkle.validations import Receipt
    >>>
    >>> s = Receipt(from_dict=r.serialize())
    >>> t = Receipt(from_json=r.toJSONString())

    or, more uniformly,

    >>> s = Receipt.deserialize(r.serialize())
    >>> t = Receipt.deserialize(r.toJSONString())

    .. note:: This is a genuine replication, since deserializations will have
        the same uuid and timestamp as the original.

    :ivar header: (*dict*) contains the keys *uuid*, *timestamp*, *validation_moment*
    :ivar body: (*dict*) contains the keys *proof_uuid*, *proof_provider*, *result*
    """

    def __init__(self, *args, **kwargs):
        """
        """
        header = {}
        body = {}
        if kwargs.get('from_dict'):                             # from json dict
            input = kwargs['from_dict']
            header.update(input['header'])
            body.update(input['body'])
        elif kwargs.get('from_json'):                           # from json text
            input = json.loads(kwargs['from_json'])
            header.update(input['header'])
            body.update(input['body'])
        else:                                                  # multiple kwargs
            header.update({
                'uuid': str(uuid.uuid1()),
                'timestamp': int(time()),
                'validation_moment': ctime(),
            })
            body.update({
                'proof_uuid': kwargs['proof_uuid'],
                'proof_provider': kwargs['proof_provider'],
                'result': kwargs['result'],
            })
        self.header = header
        self.body = body

    def __repr__(self):
        header = self.header
        body = self.body

        return '\n    ----------------------------- VALIDATION RECEIPT -----------------------------\
                \n\
                \n    uuid           : {uuid}\
                \n\
                \n    timestamp      : {timestamp} ({validation_moment})\
                \n\
                \n    proof-uuid     : {proof_uuid}\
                \n    proof-provider : {proof_provider}\
                \n\
                \n    result         : {result}\
                \n\
                \n    ------------------------------- END OF RECEIPT -------------------------------\
                \n'.format(
                    uuid=header['uuid'],
                    timestamp=header['timestamp'],
                    validation_moment=header['validation_moment'],
                    proof_uuid=body['proof_uuid'],
                    proof_provider=body['proof_provider'],
                    result='VALID' if body['result'] else 'NON VALID')

    @classmethod
    def deserialize(cls, serialized):
        """
        Deserializes the provided JSON entity

        :params serialized: a Python dict or JSON text, assumed to be the
            serialization of a *Receipt* object
        :type: dict or str
        :rtype: Receipt
        """
        kwargs = {}
        if isinstance(serialized, dict):
            kwargs.update({'from_dict': serialized})
        elif isinstance(serialized, str):
            kwargs.update({'from_json': serialized})
        return cls(**kwargs)


# Serialization

    def serialize(self):
        """
        Returns a JSON entity with the receipt's characteristics
        as key-value pairs

        :rtype: dict
        """
        return ReceiptSerializer().default(self)

    def toJSONString(self):
        """
        Returns a JSON text with the receipt's characteristics
        as key-value pairs

        :rtype: str
        """
        return json.dumps(self, cls=ReceiptSerializer, sort_keys=True, indent=4)



"""
Provides to hashing-machines their underlying encoding-machine
"""

from pymerkle.exceptions import UnsupportedEncoding, UndecodableArgumentError

ENCODINGS = ['ascii', 'big5', 'big5hkscs', 'cp037', 'cp1026', 'cp1125',
    'cp1140', 'cp1250', 'cp1251', 'cp1252', 'cp1253', 'cp1254', 'cp1255',
    'cp1256', 'cp1257', 'cp1258', 'cp273', 'cp424', 'cp437', 'cp500', 'cp775',
    'cp850', 'cp852', 'cp855', 'cp857', 'cp858', 'cp860', 'cp861', 'cp862',
    'cp863', 'cp864', 'cp865', 'cp866', 'cp869', 'cp932', 'cp949', 'cp950',
    'euc_jis_2004', 'euc_jisx0213', 'euc_jp', 'euc_kr', 'gb18030', 'gb2312',
    'gbk', 'hp_roman8', 'hz', 'iso2022_jp', 'iso2022_jp_1', 'iso2022_jp_2',
    'iso2022_jp_2004', 'iso2022_jp_3', 'iso2022_jp_ext', 'iso2022_kr',
    'iso8859_10', 'iso8859_11', 'iso8859_13', 'iso8859_14', 'iso8859_15',
    'iso8859_16', 'iso8859_2', 'iso8859_3', 'iso8859_4', 'iso8859_5',
    'iso8859_6', 'iso8859_7', 'iso8859_8', 'iso8859_9', 'johab', 'koi8_r',
    'koi8_u', 'kz1048', 'latin_1', 'mac_cyrillic', 'mac_greek', 'mac_iceland',
    'mac_latin2', 'mac_roman', 'mac_turkish', 'ptcp154', 'shift_jis',
    'shift_jis_2004', 'shift_jisx0213', 'tis_620', 'utf_16', 'utf_16_be',
    'utf_16_le', 'utf_32', 'utf_32_be', 'utf_32_le', 'utf_7', 'utf_8',]
"""Supported encoding types"""


class Encoder(object):
    """
    Encapsulates the core encoding utility of hash-machines
    """
    def __init__(self, encoding='utf-8', raw_bytes=True, security=True):
        enc = encoding.lower().replace('-', '_')
        if enc not in ENCODINGS:
            err = f'Encoding type {encoding} is not supported'
            raise UnsupportedEncoding(err)
        self.encoding  = enc
        self.raw_bytes = raw_bytes
        self.security  = security

        self.encode = self.mk_encode_func()


    def mk_encode_func(self):
        """
        Constructs and returns the core utility of the present encoding machine
        in accordance with its initial configuration (*encoding type*,
        *raw-bytes* mode and *security* mode)
        """
        encoding = self.encoding

        # Resolve security prefices

        if self.security:
            prefix_0_dec, prefix_0_enc = '\x00', bytes('\x00', encoding)
            prefix_1_dec, prefix_1_enc = '\x01', bytes('\x01', encoding)
        else:
            prefix_0_dec, prefix_0_enc = '', bytes()
            prefix_1_dec, prefix_1_enc = '', bytes()

        # Make encoding funtion

        if self.raw_bytes:
            def encode_func(left, right=None):
                if not right:
                    if isinstance(left, bytes):
                        data = prefix_0_enc + left
                    else:
                        data = prefix_0_enc + bytes(left, encoding)
                else:
                    data = prefix_1_enc + left + prefix_1_enc + right
                return data
        else:
            def encode_func(left, right=None):
                if not right:
                    if isinstance(left, bytes):
                        try:
                            left_decoded = left.decode(encoding)
                        except UnicodeDecodeError:
                            raise UndecodableArgumentError
                        data = bytes(prefix_0_dec + left_decoded,
                            encoding=encoding)
                    else:
                        data = bytes(prefix_0_dec + left, encoding)
                else:
                    try:
                        left_decoded = left.decode(encoding)
                        right_decoded = right.decode(encoding)
                    except UnicodeDecodeError:
                        raise UndecodableArgumentError
                    data = bytes(prefix_1_dec + left_decoded +
                                 prefix_1_dec + right_decoded,
                            encoding=encoding)
                return data

        return encode_func
